import numpy as np
import scipy

import eqsig
from eqsig import exceptions


def time_series_from_motion(motion, dt):
    npts = len(motion)
    return np.linspace(0, dt * (npts + 1), npts)


def determine_indices_of_peaks_for_cleaned(values):
    """
    Determines the position of values that form a local peak in a signal.

    :param values:
    :return:
    """
    diff = np.ediff1d(values, to_begin=0)
    # if negative then direction has switched
    # direction_switch = np.insert(direction_switch, 0, 0)
    peak_indices = np.where(diff[1:] * diff[:-1] < 0)[0]
    peak_indices = np.insert(peak_indices, 0, 0)  # Include first and last value
    peak_indices = np.insert(peak_indices, len(peak_indices), len(values) - 1)

    return peak_indices


# def determine_indices_of_zero_crossings_for_cleaned(values):
#     """
#     Determines the position of values that are equal or have just passed through zero.
#
#     :param values:
#     :return:
#     """
#     diff = np.diff(values)
#     # if negative then direction has switched
#     direction_switch = diff[1:] * diff[:-1]
#     direction_switch = np.insert(direction_switch, 0, 0)
#     peaks = np.where(direction_switch < 0)
#     peak_indices = peaks[0]
#     peak_indices = np.insert(peak_indices, 0, 0)  # Include first and last value
#     peak_indices = np.insert(peak_indices, len(peak_indices), len(values) - 1)
#
#     return peak_indices


def determine_peak_only_series_4_cleaned_data(values):
    """
    Determines the

    Note: array must not contain adjacent repeated values

    :param values:
    :return:
    """
    peak_indices = determine_indices_of_peaks_for_cleaned(values)
    peak_values = np.take(values, peak_indices)
    signs = np.where(np.mod(np.arange(len(peak_values)), 2), -1, 1)
    delta_peaks = np.where(-signs * peak_values < 0, -np.abs(peak_values), np.abs(peak_values))
    delta_peaks_series = np.zeros_like(values)
    np.put(delta_peaks_series, peak_indices, delta_peaks)

    return delta_peaks_series


def determine_peak_only_delta_series_4_cleaned_data(values):
    """
    Determines the

    Note: array must not contain adjacent repeated values

    :param values:
    :return:
    """
    peak_indices = determine_indices_of_peaks_for_cleaned(values)
    peak_values = np.take(values, peak_indices)
    delta_peaks = np.diff(peak_values)
    delta_peaks = np.insert(delta_peaks, 0, 0)
    delta_peaks_series = np.zeros_like(values)
    assert len(delta_peaks) == len(peak_indices)
    np.put(delta_peaks_series, peak_indices, delta_peaks)

    return delta_peaks_series


def clean_out_non_changing(values):
    """
    Takes an array removes all values that are the same as the previous value.

    :param values: array of floats
    :return: cleaned array, indices of clean values in original array
    """
    # diff_values = np.diff(values)
    # diff_values = np.insert(diff_values, 0, values[0])
    diff_values = np.ediff1d(values, to_begin=values[0])
    non_zero_indices = np.where(diff_values != 0)[0]
    non_zero_indices = np.insert(non_zero_indices, 0, 0)

    cleaned_values = np.take(values, non_zero_indices)
    return cleaned_values, non_zero_indices


def get_peak_array_indices(values, ptype='all'):
    """
    Find the indices for all of the local maxima and minima

    Parameters
    ----------
    :param values: array_like, array of values
    :return:

    Examples
    --------
    >>> values = np.array([0, 2, 1, 2, -1, 1, 1, 0.3, -1, 0.2, 1, 0.2])
    np.array([0, 2, 1, 2, -1, 1, 1, 0.3, -1, 0.2, 1, 0.2])
    >>> get_peak_array_indices(values)
    np.array([0, 1, 2, 3, 4, 5, 8, 10, 11])
    """
    # enforce array type
    values = np.array(values, dtype=float)
    # remove all non-changing values
    cleaned_values, non_zero_indices = clean_out_non_changing(values)
    # cleaned_values *= np.sign(cleaned_values[1])  # ensure first value is increasing
    peak_cleaned_indices = determine_indices_of_peaks_for_cleaned(cleaned_values)
    peak_full_indices = np.take(non_zero_indices, peak_cleaned_indices)
    if ptype == 'min':
        if values[1] - values[0] <= 0:
            return peak_full_indices[1::2]
        else:
            return peak_full_indices[::2]
    elif ptype == 'max':
        if values[1] - values[0] > 0:
            return peak_full_indices[1::2]
        else:
            return peak_full_indices[::2]
    return peak_full_indices


def get_peak_indices(asig):
    return get_peak_array_indices(asig.values)


def get_zero_crossings_array_indices(values, keep_adj_zeros=False):
    """
    Find the indices for values that are equal to zero or just passed through zero

    Parameters
    ----------
    values: array_like
        array of values
    keep_adj_zeros: bool,
        if false then if adjacent zeros are found, only first is included
    :return:

    Examples
    --------
    >>> values = np.array([0, 2, 1, 2, -1, 1, 0, 0, 1, 0.3, 0, -1, 0.2, 1, 0.2])
    np.array([0, 2, 1, 2, -1, 1, 0, 0, 1, 0.3, 0, -1, 0.2, 1, 0.2])
    >>> get_zero_crossings_array_indices(values, keep_adj_zeros=False)
    np.array([0, 4, 5, 6, 10, 12])
    """
    # enforce array type
    values = np.array(values, dtype=float)
    # get all zero values
    zero_indices = np.where(values == 0)[0]
    if not keep_adj_zeros:
        diff_is = np.ediff1d(zero_indices, to_begin=10)
        no_adj_is = np.where(diff_is > 1)[0]
        zero_indices = np.take(zero_indices, no_adj_is)
    # if negative then sign has switched
    sign_switch = values[1:] * values[:-1]
    sign_switch = np.insert(sign_switch, 0, values[0])
    through_zero_indices = np.where(sign_switch < 0)[0]
    all_zc_indices = np.concatenate((zero_indices, through_zero_indices))
    all_zc_indices.sort()
    return all_zc_indices


def get_zero_crossings_indices(asig):
    return get_zero_crossings_array_indices(asig.values)


def determine_peaks_only_delta_series(values):
    """
    Creates an array with the changes between peak values and zeros for non-peak values.

    Parameters
    ----------
    :param values: array_like, array of values
    :return:

    Examples
    --------
    >>> values = np.array([0, 2, 1, 2, 0, 1, 0, -1, 0, 1, 0])
    np.array([0, 2, 1, 2, 0.3, 1, 0.3, -1, 0.4, 1, 0])
    >>> determine_peaks_only_delta_series(values)
    array([0,  2, -1,  1,  0,  -1,  0,  -2,  0,  2,  0])
    """
    # enforce array type
    values = np.array(values)
    # rebase to zero as first value
    values -= values[0]
    # remove all non-changing values
    cleaned_values, non_zero_indices = clean_out_non_changing(values)
    cleaned_values *= np.sign(cleaned_values[1])  # ensure first value is increasing
    # compute delta peaks for cleaned data
    cleaned_delta_peak_series = determine_peak_only_delta_series_4_cleaned_data(cleaned_values)
    # re-index data to uncleaned array
    delta_peaks_series = np.zeros_like(values)
    np.put(delta_peaks_series, non_zero_indices, cleaned_delta_peak_series)
    return delta_peaks_series


def determine_pseudo_cyclic_peak_only_series(values):
    """
    Creates an array with only peak values assuming an alternative sign and zeros for non-peak values.

    Parameters
    ----------
    :param values: array_like, array of values
    :return:

    Examples
    --------
    >>> values = np.array([0, 2, 1, 2, 0, 1, 0, -1, 0, 1, 0])
    np.array([0, 2, 1, 2, 0.3, 1, 0.3, -1, 0.4, 1, 0])
    >>> determine_pseudo_cyclic_peak_only_series(values)
    array([0,  2, -1,  2,  0,  1,  0,  1,  0,  1,  0])
    """
    # enforce array type
    values = np.array(values)
    # rebase to zero as first value
    values -= values[0]
    # remove all non-changing values
    cleaned_values, non_zero_indices = clean_out_non_changing(values)
    cleaned_values *= np.sign(cleaned_values[1])  # ensure first value is increasing
    # compute delta peaks for cleaned data
    cleaned_delta_peak_series = determine_peak_only_series_4_cleaned_data(cleaned_values)
    # re-index data to uncleaned array
    delta_peaks_series = np.zeros_like(values)
    np.put(delta_peaks_series, non_zero_indices, cleaned_delta_peak_series)
    return delta_peaks_series


def fas2signal(fas, dt, stype="signal"):
    """
    Convert a fourier spectrum to time series signal

    :param fas: positive part only
    :param dt: time step of time series
    :return:
    """
    from eqsig.single import Signal, AccSignal
    n = 2 * len(fas)
    a = np.zeros(2 * len(fas), dtype=complex)
    a[1:n // 2] = fas[1:]
    a[n // 2 + 1:] = np.flip(np.conj(fas[1:]), axis=0)
    a /= dt
    s = np.fft.ifft(a)
    npts = int(2 ** (np.log(n) / np.log(2)))
    s = s[:npts]
    if stype == 'signal':
        return Signal(s, dt)
    else:
        return AccSignal(s, dt)


def generate_fa_spectrum(sig, n_pad=True):
    """
    Produces the Fourier amplitude spectrum

    Parameters
    ----------
    sig: eqsig.Signal

    Returns
    -------
    fa_spectrum: complex array_like
        Complex values of the spectrum
    fa_frequencies: array_like
        Frequencies of the spectrum
    """
    npts = sig.npts
    if n_pad:
        n_factor = 2 ** int(np.ceil(np.log2(npts)))
        fa = scipy.fft(sig.values, n=n_factor)
        points = int(n_factor / 2)
        assert len(fa) == n_factor
    else:
        fa = scipy.fft(sig.values)
        points = int(sig.npts / 2)
    fa_spectrum = fa[range(points)] * sig.dt
    fa_frequencies = np.arange(points) / (2 * points * sig.dt)
    return fa_spectrum, fa_frequencies


def interp_array_to_approx_dt(values, dt, target_dt=0.01, even=True):
    factor = dt / target_dt
    if factor == 1:
        pass
    elif factor > 1:
        factor = int(np.ceil(factor))
    else:
        factor = 1 / np.floor(1 / factor)
    t_int = np.arange(len(values))
    new_npts = factor * len(values)
    if even:
        new_npts = 2 * int(new_npts / 2)
    t_db = np.arange(new_npts) / factor
    acc_interp = np.interp(t_db, t_int, values)
    return acc_interp, dt / factor


def interp_to_approx_dt(asig, target_dt=0.01, even=True):
    """
    Interpolate a signal to a new time step

    Only a target time step is provided and the algorithm determines
     what time step is best to minimise loss of data from aliasing

    Parameters
    ----------
    asig: eqsig.AccSignal
        Acceleration time series object
    target_dt: float
        Target time step
    even: bool
        If true then forces the number of time steps to be an even number

    Returns
    -------

    """
    acc_interp, dt_interp = interp_array_to_approx_dt(asig.values, asig.dt, target_dt=target_dt, even=even)
    return eqsig.AccSignal(acc_interp, dt_interp)


def resample_to_approx_dt(asig, target_dt=0.01, even=True):
    """
    Resample a signal assuming periodic to a new time step

    Only a target time step is provided and the algorithm determines
     what time step is best to minimise loss of data from aliasing

    Parameters
    ----------
    asig: eqsig.AccSignal
        Acceleration time series object
    target_dt: float
        Target time step
    even: bool
        If true then forces the number of time steps to be an even number

    Returns
    -------

    """
    factor = asig.dt / target_dt
    if factor == 1:
        pass
    elif factor > 1:
        factor = int(np.ceil(factor))
    else:
        factor = 1 / np.floor(1 / factor)
    new_npts = factor * asig.npts
    if even:
        new_npts = 2 * int(new_npts / 2)
    acc_interp = scipy.signal.resample(asig.values, new_npts)
    return eqsig.AccSignal(acc_interp, asig.dt / factor)


def get_switched_peak_indices(asig):
    """
    Find the indices for largest peak between each zero crossing

    Parameters
    ----------
    asig: eqsig.AccSignal

    Returns
    -------
    array_like
    """
    values = asig
    if hasattr(asig, "values"):
        values = asig.values
    return get_switched_peak_array_indices(values)


def get_switched_peak_array_indices(values):
    """
    Find the indices for largest peak between each zero crossing

    Parameters
    ----------
    values: array_like

    Returns
    -------
    array_like
    """
    peak_indices = eqsig.get_peak_array_indices(values)
    peak_values = np.take(values, peak_indices)

    last = peak_values[0]
    new_peak_indices = []
    peak_values_set = [0]
    peak_indices_set = [0]
    for i in range(1, len(peak_values)):
        if peak_values[i] * last <= 0:  # only add index if sign changes (negative number)
            i_max_set = np.argmax(np.abs(peak_values_set))
            new_peak_indices.append(peak_indices_set[i_max_set])

            last = peak_values[i]
            peak_values_set = []  # reset set
            peak_indices_set = []

        peak_values_set.append(peak_values[i])
        peak_indices_set.append(i)
        
    if len(peak_values_set):  # add last
        i_max_set = np.argmax(np.abs(peak_values_set))
        new_peak_indices.append(peak_indices_set[i_max_set])
        peak_values_set.append(peak_values[i])
        peak_indices_set.append(i)

    switched_peak_indices = np.take(peak_indices, new_peak_indices)
    return switched_peak_indices


def get_sig_array_indexes_range(fas1_smooth, ratio=15):
    max_fas1 = max(fas1_smooth)
    lim_fas = max_fas1 / ratio
    indys = np.where(fas1_smooth > lim_fas)[0]
    return indys[0], indys[-1]

    # min_freq_i = 10000
    # max_freq_i = 10000
    # for i in range(len(fas1_smooth)):
    #     if fas1_smooth[i] > lim_fas:
    #         min_freq_i = i
    #         break
    # for i in range(len(fas1_smooth)):
    #     if fas1_smooth[-1 - i] > lim_fas:
    #         max_freq_i = len(fas1_smooth) - i
    #         break
    # return min_freq_i, max_freq_i


def get_sig_freq_range(asig, ratio=15):
    indices = get_sig_array_indexes_range(asig.smooth_fa_spectrum, ratio=ratio)
    return np.take(asig.smooth_fa_frequencies, indices)


def calc_fourier_moment(asig, n):
    """
    Original source unknown.

    See :cite:`Rathje:2008va`

    Parameters
    ----------
    asig
    n

    Returns
    -------

    """
    return 2 * np.trapz((2 * np.pi * asig.fa_frequencies) ** n * asig.fa_spectrum ** 2, x=asig.fa_frequencies)


def get_bandwidth_boore_2003(asig):
    m0 = calc_fourier_moment(asig, 0)
    m2 = calc_fourier_moment(asig, 2)
    m4 = calc_fourier_moment(asig, 4)
    return np.sqrt(m2 ** 2 / (m0 * m4))


def put_array_in_2d_array(values, shifts, clip='none'):
    """
    Creates a 2D array where values appear on each line, shifted by a set of indices

    Parameters
    ----------
    values: array_like (1d)
        Values to be shifted
    shifts: array_like (int)
        Indices to shift values
    clip: str or none
        if 'end' then returned 2D array trims values that overlap end of input values array

    Returns
    -------
    array_like (2D)
    """
    npts = len(values)
    # assert shifts is integer array
    end_extras = np.max([np.max(shifts), 0])
    start_extras = - np.min([np.min(shifts), 0])
    out = np.zeros((len(shifts), npts + start_extras + end_extras))
    for i, j in enumerate(shifts):
        out[i, start_extras + j:start_extras + npts + j] = values
    if clip in ['end', 'both'] and end_extras > 0:
            out = out[:, :-end_extras]
    if clip in ['start', 'both']:
        return out[:, start_extras:]
    else:
        return out


def join_sig_w_time_shift(sig, time_shifts, jtype='add'):
    """
    Zero pads values of a signal by an array of time shifts and joins it with the original

    Parameters
    ----------
    sig: eqsig.Signal
        signal to be shifted
    time_shifts: array_like
        Time shifts to be performed
    jtype: str (default='add')
        if = 'add' then shifted and original signals are added, if ='sub' then subtracted

    Returns
    -------
    shifted_values: array_like [2D shape(len(sig.values), len(shift))]

    """
    shifts = np.array(time_shifts / sig.dt, dtype=int)
    values = sig.values
    return join_values_w_shifts(values, shifts, jtype=jtype)


def join_values_w_shifts(values, shifts, jtype='add'):
    """
    Zero pads values by an array of shifts and joins it with the original values

    Parameters
    ----------
    values: array_like
        values to be shifted
    shifts: array_like [int]
        Shifts to be performed
    jtype: str (default='add')
        if = 'add' then shifted and original values are added, if ='sub' then subtracted

    Returns
    -------
    shifted_values: array_like [2D shape(len(values), len(shift))]

    """
    a0 = np.pad(values, (0, np.max(shifts)), mode='constant', constant_values=0)  # 1d
    a1 = put_array_in_2d_array(values, shifts)
    if jtype == 'add':
        return a1 + a0
    elif jtype == 'sub':
        return -a1 + a0


def get_section_average(series, start=0, end=-1, index=False):
    """
    Gets the average value of a part of series.

    Common use is so that it can be patched with another record.

    :param series: A TimeSeries object
    :param start: int or float, optional,
        Section start point
    :param end: int or float, optional,
        Section end point
    :param index: bool, optional,
        if False then start and end are considered values in time.
    :return float,
        The mean value of the section.
    """
    s_index, e_index = time_indices(series.npts, series.dt, start, end, index)

    section_average = np.mean(series.values[s_index:e_index])
    return section_average


def time_indices(npts, dt, start, end, index):
    """
    Determine the new start and end indices of the time series.

    :param npts: Number of points in original time series
    :param dt: Time step of original time series
    :param start: int or float, optional, New start point
    :param end: int or float, optional, New end point
    :param index: bool, optional, if False then start and end are considered values in time.
    :return: tuple, start index, end index
    """
    if index is False:  # Convert time values into indices
        if end != -1:
            e_index = int(end / dt) + 1
        else:
            e_index = end
        s_index = int(start / dt)
    else:
        s_index = start
        e_index = end
    if e_index > npts:
        raise exceptions.SignalProcessingWarning("Cut point is greater than time series length")
    return s_index, e_index


def generate_smooth_fa_spectrum(smooth_fa_frequencies, fa_frequencies, fa_spectrum, band=40):
    smooth_fa_spectrum = np.zeros_like(smooth_fa_frequencies)
    for i in range(smooth_fa_frequencies.size):
        f_centre = smooth_fa_frequencies[i]
        amp_array = np.log10((fa_frequencies / f_centre) ** band)

        amp_array[0] = 0

        wb_vals = np.zeros((len(amp_array)))
        for j in range(len(amp_array)):
            if amp_array[j] == 0:
                wb_vals[j] = 1
            else:
                wb_vals[j] = (np.sin(amp_array[j]) / amp_array[j]) ** 4

        smooth_fa_spectrum[i] = (sum(abs(fa_spectrum) * wb_vals) / sum(wb_vals))
    return smooth_fa_spectrum


def calc_step_fn_vals_error(values, pow=1, dir=None):
    """
    Calculates the error function generated by fitting a step function
    to the values

    Note: Assumes minimum error is at the minimum sum of the error,
    regardless of the `pow`. I.e. The best fit is taken as the mean
    of the values.

    Parameters
    ----------
    values: array_like
    pow: int
        The power that the error should be raised to
    dir: str
        Desired direction of the step function
        if 'down', then all upward steps are set to 10x maximum error
        if 'up', then all downward steps are set to 10x maximum error
        else, no modification to error

    Returns
    -------
    array_like (len same as input array)
    """
    values = np.array(values)
    npts = len(values)
    pre_a = np.tril(values, k=0)
    post_a = np.triu(values, k=0)
    pre_n = np.arange(1, len(values) + 1)
    post_n = np.arange(len(values), 0, -1)
    pre_mean = np.sum(pre_a, axis=1) / pre_n
    post_mean = np.sum(post_a, axis=1) / post_n
    err_pre = np.sum(np.abs(pre_a - pre_mean[:, np.newaxis]) ** pow, axis=1) - (npts - pre_n) * pre_mean ** pow
    err_post = np.sum(np.abs(post_a - post_mean[:, np.newaxis]) ** pow, axis=1) - (npts - post_n) * post_mean ** pow
    # case of 0 has been remove, n + 1 options
    # consider cases where it happens in between two steps
    err = np.ones_like(values)
    err[:-1] = err_post[1:] + err_pre[:-1]
    err[-1] = np.sum(np.abs(values - np.mean(values)) ** pow)
    if dir == 'down':  # if step has to be downward, then increase error for upward steps
        max_err = np.max(err)
        err = np.where(pre_mean < post_mean, max_err * 10, err)
    if dir == 'up':
        max_err = np.max(err)
        err = np.where(pre_mean > post_mean, max_err * 10, err)
    return err


def calc_step_fn_steps_vals(values, ind=None):
    if ind is None:
        ind = np.argmin(calc_step_fn_vals_error(values))
    pre = np.mean(values[:ind])
    post = np.mean(values[ind + 1:])
    return pre, post


def calc_roll_av_vals(values, steps, mode='forward'):
    """
    Calculates the rolling average of a series of values

    Parameters
    ----------
    values: array_like
    steps: int
        size of window to average over
    mode: str (default='forward')
        if 'forward' current value at start of window
        if 'backward' current value at end of window
        if 'centre' or 'center' current value in centre of window

    Returns
    -------
    array_like (len same as input array)
    """
    steps = int(steps)
    if mode == 'forward':
        x_ext = np.concatenate([values, values[-1] * np.ones(steps - 1)])
    elif mode == 'backward':
        x_ext = np.concatenate([values[0] * np.ones(steps - 1), values])
    else:
        s = int(np.floor(steps / 2))
        e = steps - s - 1
        x_ext = np.concatenate([values[0] * np.ones(s), values, values[-1] * np.ones(e)])
    csum = np.zeros(len(values) + steps)
    csum[1:] = np.cumsum(x_ext, dtype=float)

    return (csum[steps:] - csum[:-steps]) / steps


if __name__ == '__main__':
    vals = np.arange(4, 6)
    sfs = np.array([-1, 2])
    expected_full = np.array([[4, 5, 0, 0, 0],
                              [0, 0, 0, 4, 5],
                              ])
    out_a = put_array_in_2d_array(vals, sfs, clip='start')
    # out_a = join_values_w_shifts(vals, sfs, jtype='sub')

    print(out_a)



import numpy as np
import scipy
import scipy.fftpack
import scipy.signal
import eqsig.multiple


def plot_stock(splot, asig, norm_x=False, norm_all=False, interp=False, cmap=None):
    """
    Plots the Stockwell transform of an acceleration signal

    Parameters
    ----------
    splot: matplotlib.pyplot.subplot
    asig: eqsig.Signal
    norm_x: bool, default=False
        If true then all values at a time step are normalised by the maximum value
    norm_all: bool, default=False
        If true the all values are normalised by the maximum value
    Returns
    -------
        None
    """
    if not hasattr(asig, "swtf"):
        asig.swtf = transform(asig.values, interp=interp)

    b = abs(asig.swtf)
    if interp:
        b = b[int(len(b) / 2):]
    if norm_all:
        b = b / np.max(b)
    if norm_x:
        b = b / np.max(b, axis=0)
    max_freq = 1 / asig.dt / 2
    # if interp:
    #     max_freq /= 2
    min_freq = 1.0 / (len(asig.swtf) * asig.dt)

    extent = (0, asig.time[-1], min_freq, max_freq)

    kwargs = {}
    if cmap is not None:
        kwargs['cmap'] = cmap  # rainbow, gnuplot2, plasma
    splot.imshow(b, aspect='auto', extent=extent, **kwargs)


def plot_tifq_vals(subplot, tifq_vals, dt, norm_all=False, norm_x=False, cmap=None):
    """
    Plots a time frequency spectrum (e.g. Stockwell transform)

    Same as `plot_stock` but takes in values (tifq_vals) and time step (dt) instead of
    AccSignal object.

    Parameters
    ----------
    subplot: matplotlib.pyplot.subplot
    tifq_vals: 2-d array
        time-frequency values
    dt: float
        time step
    norm_x: bool, default=False
        If true then all values at a time step are normalised by the maximum value
    norm_all: bool, default=False
        If true the all values are normalised by the maximum value
    Returns
    -------
        None
    """

    b = tifq_vals
    if norm_all:
        b = b / np.max(b)
    if norm_x:
        b = b / np.max(b, axis=0)
    max_freq = 1 / dt / 2
    # if interp:
    #     max_freq /= 2
    min_freq = 1.0 / (len(tifq_vals) * dt)

    total_time = len(tifq_vals[0]) * dt
    extent = (0, total_time, min_freq, max_freq)

    kwargs = {}
    if cmap is not None:
        kwargs['cmap'] = cmap  # rainbow, gnuplot2, plasma
    subplot.imshow(b, aspect='auto', extent=extent, **kwargs)


def get_stockwell_freqs(asig):
    return (len(asig.swtf) / 2 - np.arange(0, int(len(asig.swtf) / 2))) / (len(asig.values) * asig.dt)


def get_stockwell_times(asig):
    return np.arange(0, len(asig.swtf[0])) * asig.dt / 2


def plot_fas_at_time(splot, asig, time):
    """Plots the Fourier amplitude spectrum at a time based on a Stockwell transform"""
    if not hasattr(asig, "swtf"):
        asig.swtf = transform(asig.values)
    indy = int(time / asig.dt) * 2  # 2 since Stockwell has twice as many points
    points = len(asig.swtf)
    freqs = np.arange(1, points + 1) / (points * asig.dt)
    splot.plot(freqs, np.flipud(abs(asig.swtf[:, indy])))


def plot_windowed_fas_at_time(splot, asig, time, time_window=3):
    """Plots the time averaged Fourier amplitude spectrum at a time based on a Stockwell transform"""
    if not hasattr(asig, "swtf"):
        asig.swtf = transform(asig.values)
    indy = int(time / asig.dt) * 2  # 2 since Stockwell has twice as many points
    window = int(time_window / (asig.time[-1] / asig.npts))
    s_i = max(int(indy - window / 20), 0)
    f_i = max(int(indy + window / 20), len(asig.swtf))
    points = len(asig.swtf)
    freqs = np.arange(1, points + 1) / (points * asig.dt)
    splot.plot(freqs, np.flipud(np.mean(abs(asig.swtf[:, s_i:f_i]), axis=1)))


def generate_gaussian(n_d2):
    """create a gaussian distribution"""
    f_half = np.arange(0, n_d2 + 1, 1) / (2 * n_d2)
    f = np.concatenate((f_half, np.flipud(-f_half[1:-1])))
    p = 2 * np.pi * np.outer(f, 1. / f_half[1:])
    return np.exp(-p ** 2 / 2).transpose()  # * np.exp(1j * p / n_d2)


def transform(acc, interp=False):
    """
    Performs a Stockwell transform on an array

    Assumes m = 1, p = 1

    :param acc: array_like
    :return:
    """
    # Interpolate here because function drops a time step
    if interp:
        t_int = np.arange(len(acc))
        t_db = np.arange(2 * len(acc)) / 2
        acc_db = np.interp(t_db, t_int, acc)
        n_d2 = int(len(acc))
    else:
        acc_db = acc
        n_d2 = int(len(acc) / 2)
    n_factor = 2 * n_d2
    gaussian = generate_gaussian(n_d2)
    # st0 = np.mean(acc) * np.ones(n_factor)

    fa = scipy.fftpack.fft(acc_db, n_factor, overwrite_x=True)
    diag_con = scipy.linalg.toeplitz(np.conj(fa[:n_d2 + 1]), fa)
    diag_con = diag_con[1:n_d2 + 1, :]  # first line is zero frequency
    skip_is = 0  # can skip more low frequencies since they tend to be zero
    stock = np.flipud(scipy.fftpack.ifft(diag_con[skip_is:, :] * gaussian[skip_is:, :], axis=1))

    # stock = np.insert(stock, 0, st0, 0)
    for i in range(skip_is):
        stock = np.insert(stock, 0, 0, 0)

    return stock


def itransform(stock):
    """Performs an inverse Stockwell Transform"""
    return np.real(scipy.fftpack.ifft(np.sum(stock, axis=1)))


def get_max_stockwell_freq(asig):
    if not hasattr(asig, "swtf"):
        asig.swtf = transform(asig.values)
    points = len(asig.swtf)
    freqs = np.arange(1, points + 1) / (2 * points * asig.dt)
    freqs = np.flipud(freqs)
    indy_max = np.argmax(abs(asig.swtf), axis=0)
    max_f = np.take(freqs, indy_max)
    return max_f


def get_max_tifq_vals_freq(tifq_values, dt):
    points = len(tifq_values)
    freqs = np.arange(1, points + 1) / (2 * points * dt)
    freqs = np.flipud(freqs)
    indy_max = np.argmax(abs(tifq_values), axis=0)
    max_f = np.take(freqs, indy_max)
    return max_f


def plot_max_freq_azimuth(splot, asig1, asig2, max_f=None, norm=False, r_steps=90):
    r = np.linspace(0, 180, r_steps)
    theta = np.pi * r / 180
    ys = []
    for i in range(len(r)):
        asigc = eqsig.multiple.combine_at_angle(asig1, asig2, r[i])
        freqs = np.clip(get_max_stockwell_freq(asigc), None, max_f)
        if norm:
            norm_freqs = (freqs - np.min(freqs)) / (np.max(freqs) - np.min(freqs))
            ys.append(norm_freqs)
        else:
            ys.append(freqs)

    ys = np.array(ys).T
    splot.pcolormesh(theta[np.newaxis], asig1.time[:, np.newaxis], ys)
    splot.set_thetamin(0)
    splot.set_thetamax(180)


if __name__ == '__main__':
    from tests import conftest
    from eqsig import load_signal
    import matplotlib.pyplot as plt
    import eqsig

    asig = load_signal(conftest.TEST_DATA_DIR + "test_motion_dt0p01.txt", astype="acc_sig")
    asig = eqsig.interp_to_approx_dt(asig, 0.05)
    bf, sps = plt.subplots()
    plot_stock(sps, asig, norm_x=True, interp=True, cmap='plasma')
    # , times = (10, 30), freqs = (0, 7)
    plt.show()



import numpy as np
import scipy.integrate
from eqsig import sdof
from eqsig.exceptions import deprecation


def calc_significant_duration(motion, dt, start=0.05, end=0.95):
    """
    Deprecated. Use calc_sig_dur_vals

    Parameters
    ----------
    motion
    dt
    start
    end

    Returns
    -------

    """
    deprecation("Use calc_sig_dur_vals()")
    return calc_sig_dur_vals(motion, dt, start=start, end=end)


def calc_sig_dur_vals(motion, dt, start=0.05, end=0.95, se=False):
    """
    Computes the significant duration using cumulative acceleration according to Trifunac and Brady (1975).

    Parameters
    ----------
    motion: array-like
        acceleration time series
    dt: float
        time step
    start: float, default=0.05
        threshold to start the duration
    end: float, default=0.95
        threshold to end the duration
    se: bool, default=False
        If true then return the start and end times

    Returns
    -------
    tuple (start_time, end_time)
    """

    acc2 = motion ** 2
    cum_acc2 = np.cumsum(acc2)
    ind2 = np.where((cum_acc2 > start * cum_acc2[-1]) & (cum_acc2 < end * cum_acc2[-1]))
    start_time = ind2[0][0] * dt
    end_time = ind2[0][-1] * dt

    if se:
        return start_time, end_time
    return end_time - start_time


def calc_sig_dur(asig, start=0.05, end=0.95, im=None, se=False):
    """
    Computes the significant duration according to Trifunac and Brady (1975).

    Parameters
    ----------
    asig: eqsig.AccSignal
        acceleration time series object
    dt: float
        time step
    start: float, default=0.05
        threshold to start the duration
    end: float, default=0.95
        threshold to end the duration
    im: function or None (default=None)
        A function that calculates a cumulative intensity measure, if =None, then use eqsig.im.calc_arias_intensity
    se: bool, default=False
        If true then return the start and end times

    Returns
    -------
    tuple (start_time, end_time)
    """
    if im is None:
        im_vals = calc_arias_intensity(asig)
    else:
        im_vals = im(asig)
    ind2 = np.where((im_vals > start * im_vals[-1]) & (im_vals < end * im_vals[-1]))
    start_time = ind2[0][0] * asig.dt
    end_time = ind2[0][-1] * asig.dt
    if se:
        return start_time, end_time
    return end_time - start_time


def calculate_peak(motion):
    """Calculates the peak absolute response"""
    deprecation("Use calc_peak instead of calculate_peak")
    return max(abs(min(motion)), max(motion))


def calc_peak(motion):
    """Calculates the peak absolute response"""
    return max(abs(min(motion)), max(motion))


def calc_sir(acc_sig):
    """
    Calculates the shaking intensity rate

    ref:
    Dashti, S., Bray, J. D., Pestana, J. S., Riemer, M., and Wilson, D., 2010. Centrifuge testing to
    evaluate and mitigate liquefaction-induced building settlement mechanisms,
    ASCE Journal of Geotechnical and Geoenv. Eng. 136, 918-929

    Parameters
    ----------
    acc_sig: eqsig.AccSignal
        acceleration signal

    Returns
    -------
    float
    """
    ai_total = acc_sig.arias_intensity
    t5, t75 = calc_significant_duration(acc_sig.values, acc_sig.dt)
    return 0.7 * ai_total / (t75 - t5)


def _raw_calc_arias_intensity(acc, dt):
    acc2 = acc ** 2
    return np.pi / (2 * 9.81) * scipy.integrate.cumtrapz(acc2, dx=dt, initial=0)


def calc_arias_intensity(acc_sig):
    """
    Calculates the Arias Intensity

    Parameters
    ----------
    acc_sig: eqsig.AccSignal

    Returns
    -------
    array_like
        A time series of the build up of Arias Intensity
    """

    return _raw_calc_arias_intensity(acc_sig.values, acc_sig.dt)


def calc_cav(acc_sig):
    """
    Calculates the Cumulative Absolute velocity

    ref:
    Electrical Power Research Institute. Standardization of the Cumulative
    Absolute Velocity. 1991. EPRI TR-100082-1'2, Palo Alto, California.
    """
    abs_acc = np.abs(acc_sig.values)
    return scipy.integrate.cumtrapz(abs_acc, dx=acc_sig.dt, initial=0)


def calc_cav_dp(asig):
    """
    Calculates standardized cumulative absolute velocity

    ref:
    Campbell KW, Bozorgnia Y. Predictive equations for the horizontal component of standardized
    cumulative absolute velocity as adapted for use in the shutdown of U.S. nuclear power plants.
    Nucl Eng Des 2011;241:2558-69.

    :param asig:
    :return:
    """
    start = 0
    pga_max = 0
    cav_dp = 0
    points_per_sec = (int(1 / asig.dt))
    total_seconds = int(asig.time[-1])
    cav_dp_1_series = []
    acc_in_g = asig.values / 9.81

    for i in range(0, total_seconds):

        end = start + points_per_sec

        interval_total_time = (start * asig.dt) + 1
        interval_time = np.arange(start * asig.dt, interval_total_time, asig.dt)

        acc_interval = []
        for j in range(start, end + 1):
            acc_interval.append(acc_in_g[j])

        acc_interval = np.array(acc_interval)
        abs_acc_interval = abs(acc_interval)

        x_lower = start * asig.dt  # the lower limit of x
        x_upper = end * asig.dt  # the upper limit of x
        x_int = interval_time[np.where((x_lower <= interval_time) * (interval_time <= x_upper))]
        y_int = np.abs(np.array(abs_acc_interval)[np.where((x_lower <= interval_time) * (interval_time <= x_upper))])
        int_acc = scipy.integrate.trapz(y_int, x_int)

        # calculation of pga (g)
        pga = (max(abs_acc_interval))
        if pga > pga_max:
            pga_max = pga

        if (pga - 0.025) < 0:
            h = 0
        elif (pga - 0.025) >= 0:
            h = 1
        else:
            raise ValueError("cannot evaluate pga: {0}".format(pga))

        cav_dp = cav_dp + (h * int_acc)
        cav_dp_1_series.append(cav_dp)
        start = end
    t1s = np.arange(total_seconds)
    cav_dp_time_series = np.interp(asig.time, t1s, cav_dp_1_series)
    return cav_dp_time_series


def calc_isv(acc_sig):
    """
    Calculates the integral of the squared velocity

    See Kokusho (2013)
    :return:
    """
    return scipy.integrate.cumtrapz(acc_sig.velocity ** 2, dx=acc_sig.dt, initial=0)


def cumulative_response_spectra(acc_signal, fun_name, periods=None, xi=None):

    if periods is None:
        periods = acc_signal.response_times
    else:
        periods = np.array(periods)
    if xi is None:
        xi = 0.05
    resp_u, resp_v, resp_a = sdof.response_series(acc_signal.values, acc_signal.dt, periods, xi)
    if fun_name == "arias_intensity":
        rs = _raw_calc_arias_intensity(resp_a, acc_signal.dt)
    else:
        raise ValueError
    return rs


def calc_max_velocity_period(asig):
    from eqsig import AccSignal
    periods = np.logspace(-1, 0.3, 100)
    new_sig = AccSignal(asig.values, asig.dt)
    new_sig.generate_response_spectrum(response_times=periods, xi=0.15)

    max_index = np.argmax(new_sig.s_v)
    max_period = periods[max_index]
    return max_period


def max_acceleration_period(asig):
    from eqsig import AccSignal
    periods = np.logspace(-1, 1, 100)
    new_sig = AccSignal(asig.values, asig.dt)
    new_sig.generate_response_spectrum(response_times=periods, xi=0)

    max_index = np.argmax(new_sig.s_a)
    max_period = periods[max_index]
    return max_period


def max_fa_period(asig):
    """Calculates the period corresponding to the maximum value in the Fourier amplitude spectrum"""
    max_index = np.argmax(asig.fa_spectrum)
    max_period = 1. / asig.fa_frequencies[max_index]
    return max_period


def calc_bandwidth_f_min(asig, ratio=0.707):
    fas1_smooth = asig.smooth_fa_spectrum
    max_fas1 = max(fas1_smooth)
    lim_fas = max_fas1 * ratio
    ind2 = np.where(fas1_smooth > lim_fas)
    min_freq = asig.smooth_fa_frequencies[ind2[0][0]]
    return min_freq


def calc_bandwidth_f_max(asig, ratio=0.707):
    fas1_smooth = asig.smooth_fa_spectrum
    max_fas1 = max(fas1_smooth)
    lim_fas = max_fas1 * ratio
    ind2 = np.where(fas1_smooth > lim_fas)
    max_freq = asig.smooth_fa_frequencies[ind2[0][-1]]
    return max_freq


def calc_bracketed_duration(asig, threshold):
    deprecation("Use calc_brac_dur")
    return calc_brac_dur(asig, threshold)


def calc_brac_dur(asig, threshold, se=False):
    """
    Calculates the Bracketed duration based on some threshold

    Parameters
    ----------
    asig: eqsig.AccSignal
        Acceleration time series object
    threshold: float
        acceleration threshold to calculation duration start and end
    se: bool, default=False
        If true then return the start and end times
    Returns
    -------
    float
    """
    abs_motion = abs(asig.values)

    time = np.arange(asig.npts) * asig.dt
    # Bracketed duration
    ind01 = np.where(abs_motion > threshold)
    time2 = time[ind01]
    try:
        if se:
            return time2[0], time2[-1]
        return time2[-1] - time2[0]
    except IndexError:
        if se:
            return None, None
        return 0



def calc_a_rms(asig, threshold):
    abs_motion = abs(asig.values)
    # Bracketed duration
    ind01 = np.where(abs_motion / 9.8 > threshold)
    try:
        # rms acceleration in m/s/s
        a_rms01 = np.sqrt(1 / asig.t_b01 * np.trapz((asig.values[ind01[0][0]:ind01[0][-1]]) ** 2, dx=asig.dt))
    except IndexError:
        a_rms01 = 0
    return a_rms01


def calc_integral_of_abs_velocity(asig):
    abs_vel = abs(asig.velocity)
    vel_int = np.cumsum(abs_vel * asig.dt)
    return vel_int


def calc_cumulative_abs_displacement(asig):
    return calc_integral_of_abs_velocity(asig)


def calc_integral_of_abs_acceleration(asig):
    abs_acc = abs(asig.values)
    acc_int = np.cumsum(abs_acc * asig.dt)
    return acc_int



import numpy as np


def single_elastic_response(motion, step, period, xi):
    """
    Perform Duhamels integral to get the displacement.
    http://www.civil.utah.edu/~bartlett/CVEEN7330/Duhamel%27s_integral.pdf
    http://www1.aucegypt.edu/faculty/mharafa/MENG%20475/Forced%20Vibration.pdf
    :param motion: acceleration in m/s2
    :param step: the time step
    :param period: The period of SDOF oscillator
    :param xi: fraction of critical damping (e.g. 0.05)
    :return:
    """
    w_n = (2.0 * np.pi) / period
    w_d = w_n * np.sqrt(1 - xi ** 2)
    x_w_n = xi * w_n
    length = len(motion)

    time = step * np.arange(length + 1)
    disp = np.zeros(length)

    p = motion * step / w_d

    for i in range(length):
        dtn = time[:-i - 1]
        d_new = p[i] * np.exp(-x_w_n * dtn) * np.sin(w_d * dtn)

        disp[i:] += d_new

    return disp


def slow_response_spectra(motion, step, periods, xis):
    """
    Perform Duhamels integral to get the displacement.
    http://www.civil.utah.edu/~bartlett/CVEEN7330/Duhamel%27s_integral.pdf
    http://www1.aucegypt.edu/faculty/mharafa/MENG%20475/Forced%20Vibration.pdf

    :param motion: acceleration in m/s2
    :param step: the time step
    :param period: The period of SDOF oscilator
    :param xi: fraction of critical damping (e.g. 0.05)
    :return:
    """
    points = len(periods)
    xi = xis[0]
    s_d = np.zeros(points)

    for i in range(points):
        s_d[i] = max(abs(single_elastic_response(motion, step, periods[i], xi)))

    s_v = s_d * 2 * np.pi / periods
    s_a = s_d * (2 * np.pi / periods) ** 2

    return s_d, s_v, s_a


def compute_a_and_b(xi, w, dt):
    """
    From the paper by Nigam and Jennings (1968), computes the two matrices.

    :param xi: critical damping ratio
    :param w: angular frequencies
    :param dt: time step
    :return: matrices A and B
    """

    # Reduce the terms since all is matrix multiplication.
    xi2 = xi * xi
    w2 = w ** 2
    one_ov_w2 = 1. / w2
    sqrt_b2 = np.sqrt(1. - xi2)
    w_sqrt_b2 = w * sqrt_b2

    exp_b = np.exp(-xi * w * dt)
    two_b_ov_w2 = (2 * xi ** 2 - 1) / (w ** 2 * dt)
    two_b_ov_w3 = 2 * xi / (w ** 3 * dt)

    sin_wsqrt = np.sin(w_sqrt_b2 * dt)
    cos_wsqrt = np.cos(w_sqrt_b2 * dt)

    # A matrix
    a_11 = exp_b * (xi / sqrt_b2 * sin_wsqrt + cos_wsqrt)  # Eq 2.7d(1)
    a_12 = exp_b / (w * sqrt_b2) * sin_wsqrt  # Eq 2.7d(2)
    a_21 = -w / sqrt_b2 * exp_b * sin_wsqrt    # Eq 2.7d(3)
    a_22 = exp_b * (cos_wsqrt - xi / sqrt_b2 * sin_wsqrt)  # Eq 2.7d(4)

    a = np.array([[a_11, a_12], [a_21, a_22]])

    # B matrix
    bsqrd_ov_w2_p_xi_ov_w = two_b_ov_w2 + xi / w
    sin_ov_wsqrt = sin_wsqrt / w_sqrt_b2
    xwcos = xi * w * cos_wsqrt
    wsqrtsin = w_sqrt_b2 * sin_wsqrt

    # Eq 2.7e
    b_11 = exp_b * (bsqrd_ov_w2_p_xi_ov_w * sin_ov_wsqrt + (two_b_ov_w3 + one_ov_w2) * cos_wsqrt) - two_b_ov_w3
    b_12 = -exp_b * (two_b_ov_w2 * sin_ov_wsqrt + two_b_ov_w3 * cos_wsqrt) - one_ov_w2 + two_b_ov_w3
    b_21 = exp_b * (bsqrd_ov_w2_p_xi_ov_w * (cos_wsqrt - xi / sqrt_b2 * sin_wsqrt)
                    - (two_b_ov_w3 + one_ov_w2) * (wsqrtsin + xwcos)) + one_ov_w2 / dt
    b_22 = -exp_b * (two_b_ov_w2 * (cos_wsqrt - xi / sqrt_b2 * sin_wsqrt) - two_b_ov_w3 * (wsqrtsin + xwcos)) - one_ov_w2 / dt

    b = np.array([[-b_11, b_12], [-b_21, -b_22]])  # Why are the signs reversed??

    return a, b


def nigam_and_jennings_response(acc, dt, periods, xi):
    """
    Implementation of the response spectrum calculation from Nigam and Jennings (1968).

    Ref: Nigam, N. C., Jennings, P. C. (1968) Digital calculation of response spectra from strong-motion earthquake
    records. National Science Foundation.

    :param acc: acceleration in m/s2
    :param periods: response periods of interest
    :param dt: time step of the acceleration time series
    :param xi: critical damping factor
    :return: response displacement, response velocity, response acceleration
    """

    acc = np.array(acc).astype(np.float)
    periods = np.array(periods).astype(np.float)
    w = 6.2831853 / periods
    dt = np.float(dt)
    xi = np.float(xi)

    # implement: delta_t should be less than period / 20

    a, b = compute_a_and_b(xi, w, dt)

    resp_u = np.zeros([len(w), len(acc)], dtype=np.float)
    resp_v = np.zeros([len(w), len(acc)], dtype=np.float)

    for i in range(len(acc) - 1):  # possibly speed up using scipy.signal.lfilter
        # x_i+1 = A cross (u, v) + B cross (acc_i, acc_i+1)  # Eq 2.7a
        resp_u[:, i + 1] = (a[0][0] * resp_u[:, i] + a[0][1] * resp_v[:, i] + b[0][0] * acc[i] + b[0][1] * acc[i + 1])
        resp_v[:, i + 1] = (a[1][0] * resp_u[:, i] + a[1][1] * resp_v[:, i] + b[1][0] * acc[i] + b[1][1] * acc[i + 1])

    w2 = w ** 2
    sdof_acc = -2 * xi * w[:, np.newaxis] * resp_v - w2[:, np.newaxis] * resp_u

    return resp_u, resp_v, sdof_acc


def absmax(a, axis=None):
    amax = a.max(axis)
    amin = a.min(axis)
    return abs(np.where(-amin > amax, amin, amax))


def pseudo_response_spectra(motion, dt, periods, xi):
    """
    Computes the maximum response displacement, pseudo velocity and pseudo acceleration.

    :param motion: array floats, acceleration in m/s2
    :param dt: float, the time step
    :param periods: array floats, The period of SDOF oscilator
    :param xi: float, fraction of critical damping (e.g. 0.05)
    :return: tuple floats, (spectral displacement, pseudo spectral velocity, pseudo spectral acceleration)
    """
    periods = np.array(periods)
    resp_u, resp_v, resp_a = nigam_and_jennings_response(motion, dt, periods, xi)
    sds = absmax(resp_u, axis=1)
    svs = 2 * np.pi / periods * sds
    sas = (2 * np.pi / periods) ** 2 * sds
    sas = np.where(periods < dt * 6, absmax(motion), sas)
    return sds, svs, sas


def response_series(motion, dt, periods, xi):
    """
    Computes the elastic response to the acceleration time series

    :param motion: array floats, acceleration in m/s2
    :param dt: float, the time step
    :param periods: array floats, The period of SDOF oscilator
    :param xi: float, fraction of critical damping (e.g. 0.05)
    :return: tuple of float arrays, (response displacements, response velocities, response accelerations)
    """
    return nigam_and_jennings_response(motion, dt, periods, xi)


def true_response_spectra(motion, dt, periods, xi):
    """
    Computes the actual maximum response values, not the pseudo values

    :param motion: array floats, acceleration in m/s2
    :param dt: float, the time step
    :param periods: array floats, The period of SDOF oscilator
    :param xi: float, fraction of critical damping (e.g. 0.05)
    :return: tuple floats, (spectral displacement, spectral velocity, spectral acceleration)
    """
    resp_u, resp_v, resp_a = nigam_and_jennings_response(motion, dt, periods, xi)
    sas = absmax(resp_a, axis=1)
    svs = absmax(resp_v, axis=1)
    sds = absmax(resp_u, axis=1)
    sas = np.where(periods < dt * 6, absmax(motion), sas)
    return sds, svs, sas


# def plot_response_spectra():
#     import matplotlib.pyplot as plt
#     step = 0.01
#     xis = [0.05]
#     periods = np.arange(1, 5, 0.5)
#     motion = np.sin(0.1 * np.arange(1000)) * 0.01
#     s_d, s_v, s_a = response_spectra(motion, step, periods, xis)
#
#     plt.plot(periods, s_a)
#     plt.show()
#
#
def time_the_generation_of_response_spectra():
    step = 0.01
    xi = 0.05
    periods = np.linspace(1, 5, 500)
    # periods = np.array([0.01])
    motion = np.sin(0.1 * np.arange(100000)) * 0.01
    # s_d, s_v, s_a = all_at_once_response_spectra(values, step, periods, xis)
    s_d, s_v, s_a = pseudo_response_spectra(motion, step, periods, xi)


if __name__ == '__main__':

    # time_response_spectra()
    time_the_generation_of_response_spectra()
    # import cProfile
    # cProfile.run('time_the_generation_of_response_spectra()')



__author__ = 'maximmillen'

from collections import OrderedDict
# from sortedcontainers import SortedDict
import numpy as np

from eqsig.single import Signal, AccSignal


class Cluster(object):

    def __init__(self, values, dt, names=[], master_index=0, stypes="custom", **kwargs):

        self.freq_range = np.array(kwargs.get('freq_range', [0.1, 20]))
        lvt = np.log10(1.0 / self.freq_range)
        if stypes == "custom" or stypes == "acc":
            stypes = [stypes] * len(values)
        self.response_times = kwargs.get('resp_times', np.logspace(lvt[1], lvt[0], 31, base=10))
        if master_index < 0:
            raise ValueError("master_index must be positive")
        if master_index > len(values) - 1:
            raise ValueError("master_index: %i, out of bounds, maximum value: %i" % (master_index, len(signals) - 1))
        self.master_index = master_index
        # if len(signals) != len(steps):
        #     raise ValueError("Length of signals: %i, must match length of steps: %i" % (len(signals), len(steps)))

        # self.master = Record(master_motion, master_step, response_times=self.response_times)
        shortage = len(values) - len(names)
        self.names = list(names)
        for i in range(shortage):
            self.names.append("m%i" % (len(names) + i))
        self.master = self.names[master_index]
        self.dt = dt
        self.signals = OrderedDict()
        for s in range(len(values)):
            if stypes[s] == "acc":
                self.signals[self.names[s]] = AccSignal(values[s], dt, self.names[s], response_times=self.response_times)
            else:
                self.signals[self.names[s]] = Signal(values[s], dt, self.names[s])

    def values_by_index(self, index):
        key_value_pair = list(self.signals.items())[index]
        return key_value_pair[1].values

    def signal_by_index(self, index):
        key_value_pair = list(self.signals.items())[index]
        return key_value_pair[1]

    def values(self, name):
        return self.signals[name].values

    def name_by_index(self, index):
        key_value_pair = list(self.signals.items())[index]
        return key_value_pair[0]

    @property
    def time(self):
        sig0 = self.signal_by_index(0)
        return sig0.time

    def same_start(self, **kwargs):
        """
        Starts both time series at the same y-value
        """
        base = kwargs.get('base', 0)
        start = kwargs.get('start', 0)
        end = kwargs.get('end', 1)
        verbose = kwargs.get('verbose', 0)
        master_average = self.signal_by_index(self.master_index).get_section_average(start=start, end=end)
        for i in range(len(self.signals)):
            if i != self.master_index:
                slave_signal = self.signal_by_index(1)
                slave_average = slave_signal.get_section_average(start=start, end=end)
                diff = slave_average - master_average
                if verbose:
                    print('Same start the records')
                    print('old difference in starts: ', diff)
                slave_signal.reset_values(slave_signal.values - diff)

    def combine_motions(self, f_ch, low_index=0, high_index=1, **kwargs):
        """
        This method combined the two ground motions by taking
        the high frequency of one values
        and the low frequency of the other.
        WARNING: records must have the same time step!!!

        Parameters
        ----------
        :param f_ch: is the frequency change point
        :param low_index: --
        :param high_index: --
        :param order: refers to the order of the BW filter
        combined values is returned as self.combo
        """
        order = kwargs.get('order', 4)
        remove_gibbs = kwargs.get('remove_gibbs', 0)
        self.signal_by_index(high_index).butter_pass(cut_off=(f_ch, None), order=order, remove_gibbs=remove_gibbs)
        self.signal_by_index(low_index).butter_pass(cut_off=(None, f_ch), order=order, remove_gibbs=remove_gibbs)
        motion = self.signal_by_index(low_index).values + self.signal_by_index(high_index).values
        return motion

    def time_match(self, **kwargs):
        """
        This method determines the time lag between two records and removes the lag
        - the solution is based on a sum of the squared residuals.
        - The motions are both set to the base values, default = 0
        - must have same time step and same magnitude
        """
        verbose = kwargs.get('verbose', 0)
        steps = kwargs.get('steps', 10)
        set_step = kwargs.get('set_step', False)
        trim = kwargs.get('trim', True)

        if set_step is False:
            if verbose:
                print('length m0: ', self.signal_by_index(0).npts)
                print('length m1: ', self.signal_by_index(1).npts)
            length_check = min(self.signal_by_index(0).npts, self.signal_by_index(1).npts)
            bm = self.signal_by_index(self.master_index).values[:length_check]
            for s in range(len(self.signals)):
                if s != self.master_index:
                    slave_signal = self.signal_by_index(s)
                    om = slave_signal.values[:length_check]
                    squares = (bm[0:-steps] - om[0:-steps]) ** 2
                    min_diff = np.sum(squares)
                    min_ind = 0
                    if verbose:
                            print('mi: ', min_ind, ' min_diff: ', min_diff)
                else:
                    continue
                # Check other values lags base values
                for i in range(steps):
                    squares = (om[i:-steps + i] - bm[0:-steps]) ** 2
                    diff = sum(squares)
                    if verbose:
                        print('ind: ', i, ' diff: ', diff)
                    if diff < min_diff:
                        min_diff = diff
                        min_ind = i + 0
                # Check base values lags other values
                for i in range(steps):
                    squares = (bm[i:-steps + i] - om[0:-steps]) ** 2
                    diff = sum(squares)
                    if verbose:
                        print('ind: ', i, ' diff: ', diff)
                    if diff < min_diff:
                        min_diff = diff
                        min_ind = -i - 0
                if verbose:
                    print('lag index: ', min_ind)

                if min_ind < 0:  # pad with initial value
                    m_temp = [om[0]] * abs(min_ind) + list(om[:min_ind])
                elif min_ind > 0:  # pad with final value
                    m_temp = list(om[min_ind:]) + [om[-1]] * abs(min_ind)
                else:
                    continue
                slave_signal.reset_values(m_temp)


        # else:
        #     min_ind = set_step
        # if verbose:
        #     print('time step lag: ', min_ind)

        # pad other record with constant
        # if verbose:
        #     print('m_temp[0]: ', m_temp[0])
        # if min_ind != 0:
        #     m = self.motions([np.mod(base + 1, 2)])
        #     m = np.ones(len(self.motions[np.mod(base + 1, 2)]))
        # if min_ind < 0:
        #     self.motions[np.mod(base + 1, 2)] = m_temp[0] * self.motions[np.mod(base + 1, 2)]
        #     self.motions[np.mod(base + 1, 2)][-min_ind:] = m_temp
        # elif min_ind > 0:
        #     self.motions[np.mod(base + 1, 2)] = m_temp[-1] * self.motions[np.mod(base + 1, 2)]
        #     self.motions[np.mod(base + 1, 2)][:-min_ind] = m_temp
        # if trim:
        #     tmax = len(self.motions[base]) * self.dts[base]
        #     ender = int(tmax / self.dts[np.mod(base + 1, 2)])
        #     if verbose:
        #         print('original length: ', len(self.motions[np.mod(base + 1, 2)]))
        #         print('trim length: ', ender)
        #     self.motions[np.mod(base + 1, 2)] = self.motions[np.mod(base + 1, 2)][:ender]
        return min_ind

    def generate_response_spectrums(self):
        for i in range(len(self.signals)):
            self.signal_by_index(i).generate_response_spectrum()

    def calculate_ratios(self):

        self.generate_response_spectrums()

        self.resp_log_ratio = np.log10(self.motions(1).s_a / self.motions(0).s_a)
        self.fa_log_ratio = np.log10(self.motions(1).smooth_fa_spectrum / self.motions(0).smooth_fa_spectrum)
        centre_freq = 1.0 / self.response_times[len(self.response_times) / 2]

        self.response_ratio_moment = sum(self.resp_log_ratio * (np.log10((1.0 /
                                         self.response_times) / centre_freq)))


def combine_at_angle(acc_sig_ns, acc_sig_we, angle):
    off_rad = np.radians(angle)
    combo = acc_sig_ns.values * np.cos(off_rad) + acc_sig_we.values * np.sin(off_rad)
    new_sig = AccSignal(combo, acc_sig_ns.dt)
    return new_sig


def compute_rotated(acc_sig_ns, acc_sig_we, angle_off_ns=0.0, parameter=None, func=None, points=100):
    """
    Computes the rotated value of a parameter.

    :param acc_sig_ns:
    :param acc_sig_we:
    :param angle_off_ns: Angle from North in degrees of the primary signal.
    :return: tuple, (angle to rotate, rotated)
    """
    assert isinstance(acc_sig_ns, AccSignal)
    assert isinstance(acc_sig_we, AccSignal)
    assert acc_sig_ns.dt == acc_sig_we.dt
    assert acc_sig_ns.npts == acc_sig_we.npts, (acc_sig_ns.npts, acc_sig_we.npts)

    degrees = np.linspace(0 - angle_off_ns, 180. - angle_off_ns, points)
    degrees = np.mod(degrees, 360)
    pvalues = []
    for i in range(len(degrees)):
        new_sig = combine_at_angle(acc_sig_ns, acc_sig_we, degrees[i])
        if parameter == "arias_intensity":
            new_sig.generate_all_motion_stats()
        if parameter is not None:
            assert func is None
            pvalues.append(getattr(new_sig, parameter))
        elif func is not None:
            val = func(new_sig)
            if hasattr(val, "__len__"):
                pvalues.append(val[-1])
            else:
                pvalues.append(val)
        else:
            raise ValueError("parameter or func must be not None")

    return degrees, np.array(pvalues)




import numpy as np
import scipy.signal as ss
import scipy

from eqsig import exceptions
from eqsig.functions import get_section_average, generate_smooth_fa_spectrum, interp_array_to_approx_dt
import eqsig.sdof as dh
import eqsig.displacements as sd
import eqsig.im as sm
from eqsig import im
from eqsig.exceptions import deprecation


class Signal(object):
    _npts = None
    _smooth_freq_points = 61
    _fa_spectrum = None
    _fa_frequencies = None
    _cached_fa = False
    _cached_smooth_fa = False

    def __init__(self, values, dt, label='m1', smooth_freq_range=(0.1, 30), verbose=0, ccbox=0):
        """
        A time series object

        Parameters
        ----------
        values: array_like
            values of the time series
        dt: float
            the time step
        label: str, optional
            A name for the signal
        smooth_freq_range: tuple, optional
            The frequency range for computing the smooth FAS
        verbose: int, optional
            Level of console output
        ccbox: int, optional
            colour id for plotting
        :return:
        """
        self.verbose = verbose
        self._dt = dt
        self._values = np.array(values)
        self.label = label
        self._smooth_freq_range = smooth_freq_range
        self._npts = len(self.values)
        self._smooth_fa_spectrum = np.zeros((self.smooth_freq_points))
        self._smooth_fa_frequencies = np.zeros(self.smooth_freq_points)
        self.ccbox = ccbox

    @property
    def values(self):
        return self._values

    def reset_values(self, new_values):
        self._values = new_values
        self._npts = len(new_values)
        self.clear_cache()

    @property
    def dt(self):
        """ The time step """
        return self._dt

    @property
    def npts(self):  # Deliberately no public setter method for this
        """ The number of points in the time series """
        return self._npts

    @property
    def time(self):
        """ An array of time of equal length to the time series """
        return np.arange(0, self.npts) * self.dt

    @property
    def fa_spectrum(self):
        """Generate the one-sided Fourier Amplitude spectrum"""
        if not self._cached_fa:
            self.generate_fa_spectrum()
        return self._fa_spectrum

    def generate_fa_spectrum(self):
        n_factor = 2 ** int(np.ceil(np.log2(self.npts)))
        fa = scipy.fft(self.values, n=n_factor)
        points = int(n_factor / 2)
        self._fa_spectrum = fa[range(points)] * self.dt
        self._fa_frequencies = np.arange(points) / (2 * points * self.dt)
        self._cached_fa = True

    @property
    def fa_frequencies(self):
        if not self._cached_fa:
            self.generate_fa_spectrum()
        return self._fa_frequencies

    @property
    def smooth_freq_range(self):
        return self._smooth_freq_range

    @smooth_freq_range.setter
    def smooth_freq_range(self, limits):
        self._smooth_freq_range = np.array(limits)
        self._cached_smooth_fa = False

    @property
    def smooth_freq_points(self):
        return self._smooth_freq_points

    @smooth_freq_points.setter
    def smooth_freq_points(self, value):
        self._smooth_freq_points = value
        self._cached_smooth_fa = False

    @property
    def smooth_fa_frequencies(self):
        if not self._cached_smooth_fa:
            self.generate_smooth_fa_spectrum()
        return self._smooth_fa_frequencies

    @property
    def smooth_fa_spectrum(self):
        if not self._cached_smooth_fa:
            self.generate_smooth_fa_spectrum()
        return self._smooth_fa_spectrum

    def clear_cache(self):
        """Resets the dynamically calculated properties."""
        self._cached_smooth_fa = False
        self._cached_fa = False

    def generate_smooth_fa_spectrum(self, band=40):
        """
        Calculates the smoothed Fourier Amplitude Spectrum
        using the method by Konno and Ohmachi 1998

        Parameters
        ----------
        band: int
            range to smooth over
        """
        lf = np.log10(self.smooth_freq_range)

        fa_frequencies = self.fa_frequencies
        fa_spectrum = self.fa_spectrum
        smooth_fa_frequencies = np.logspace(lf[0], lf[1], self.smooth_freq_points, base=10)
        self._smooth_fa_spectrum = generate_smooth_fa_spectrum(smooth_fa_frequencies, fa_frequencies,
                                                               fa_spectrum, band=band)
        self._smooth_fa_frequencies = smooth_fa_frequencies
        self._cached_smooth_fa = True

    def butter_pass(self, cut_off=(0.1, 15), **kwargs):
        """
        Performs a Butterworth filter

        Notes
        -----
        Wraps the scipy 'butter' filter

        Parameters
        ----------
        cut_off: tuple
            Lower and upper cut off frequencies for the filter, if None then no filter.
            e.g. (None, 15) applies a lowpass filter at 15Hz, whereas (0.1, 10) applies
            a bandpass filter at 0.1Hz to 10Hz.
        filter_order: int
            Order of the Butterworth filter
        remove_gibbs: str (default=None)
            Pads with zeros to remove the Gibbs filter effect,
            if = 'start' then pads at start,
            if = 'end' then pads at end,
            if = 'mid' then pads half at start and half at end
        gibbs_extra: int
            each increment of the value doubles the record length using zero padding.
        :return:
        """
        if isinstance(cut_off, list) or isinstance(cut_off, tuple) or isinstance(cut_off, np.Array):
            pass
        else:
            raise ValueError("cut_off must be list, tuple or array.")
        if len(cut_off) != 2:
            raise ValueError("cut_off must be length 2.")
        if cut_off[0] is not None and cut_off[1] is not None:
            filter_type = "band"
            cut_off = np.array(cut_off)
        elif cut_off[0] is None:
            filter_type = 'low'
            cut_off = cut_off[1]
        else:
            filter_type = 'high'
            cut_off = cut_off[0]

        filter_order = kwargs.get('filter_order', 4)
        remove_gibbs = kwargs.get('remove_gibbs', None)
        gibbs_extra = kwargs.get('gibbs_extra', 1)
        gibbs_range = kwargs.get('gibbs_range', 50)
        sampling_rate = 1.0 / self.dt
        nyq = sampling_rate * 0.5

        mote = self.values
        org_len = len(mote)

        if remove_gibbs is not None:
            # Pad end of record with extra zeros then cut it off after filtering
            nindex = int(np.ceil(np.log2(len(mote)))) + gibbs_extra
            new_len = 2 ** nindex
            diff_len = new_len - org_len
            if remove_gibbs == 'start':
                s_len = 0
                f_len = s_len + org_len
            elif remove_gibbs == 'end':
                s_len = diff_len
                f_len = s_len + org_len
            else:
                s_len = int(diff_len / 2)
                f_len = s_len + org_len

            end_value = np.mean(mote[-gibbs_range:])
            start_value = np.mean(mote[:gibbs_range])
            temp = start_value * np.ones(new_len)
            temp[f_len:] = end_value
            temp[s_len:f_len] = mote
            mote = temp
        else:
            s_len = 0
            f_len = org_len

        wp = cut_off / nyq
        b, a = ss.butter(filter_order, wp, btype=filter_type)
        mote = ss.filtfilt(b, a, mote)
        # removing extra zeros from gibbs effect
        mote = mote[s_len:f_len]  # TODO: don't use -1

        self.reset_values(mote)

    def remove_average(self, section=-1, verbose=-1):
        """
        Calculates the average and removes it from the record
        """

        if verbose == -1:
            verbose = self.verbose

        average = np.mean(self.values[:section])
        self.reset_values(self.values - average)

        if verbose:
            print('removed av.: ', average)

    def remove_poly(self, poly_fit=0):
        """
        Calculates best fit polynomial and removes it from the record
        """

        x = np.linspace(0, 1.0, self.npts)
        cofs = np.polyfit(x, self.values, poly_fit)
        y_cor = 0 * x
        for co in range(len(cofs)):
            mods = x ** (poly_fit - co)
            y_cor += cofs[co] * mods

        self.reset_values(self.values - y_cor)

    def get_section_average(self, start=0, end=-1, index=False):
        """
        Gets the average value of a part of series.

        Common use is so that it can be patched
        with another record.

        Parameters
        ----------
        start: int or float, optional
            Section start point
        end: int or float, optional
            Section end point
        index: bool, optional
            if False then start and end are considered values in time.

        Returns
        -------
        float
            The mean value of the section.
        """
        return get_section_average(self, start=start, end=end, index=index)

    def add_constant(self, constant):
        """
        Adds a single value to every value in the signal.

        :param constant:
        :return:
        """
        self.reset_values(self.values + constant)

    def add_series(self, series):
        """
        Adds a series of values to the values in the signal.

        :param series: A series of values
        :return:
        """
        if len(series) == self.npts:
            self.reset_values(self.values + series)
        else:
            raise exceptions.SignalProcessingError("new series has different length to Signal")

    def add_signal(self, new_signal):
        """
        Combines a signal.

        Parameters
        ----------
        new_signal: Signal object
        """
        if isinstance(new_signal, Signal):
            if new_signal.dt == self.dt:
                self.add_series(new_signal.values)
            else:
                raise exceptions.SignalProcessingError("New signal has different time step")
        else:
            raise exceptions.SignalProcessingError("New signal is not a Signal object")

    def running_average(self, width=1):
        """
        Averages the values over a width (chunk of numbers)
        replaces the value in the centre of the chunk.

        :param width: The range over which values are averaged
        """

        mot = self.values

        for i in range(len(mot)):
            if i < width / 2:
                cc = i + int(width / 2) + 1
                self._values[i] = np.mean(mot[:cc])
            elif i > len(mot) - width / 2:
                cc = i - int(width / 2)
                self._values[i] = np.mean(mot[cc:])
            else:
                cc1 = i - int(width / 2)
                cc2 = i + int(width / 2) + 1
                self._values[i] = np.mean(mot[cc1:cc2])

        self.clear_cache()


class AccSignal(Signal):

    def __init__(self, values, dt, label='m1', smooth_freq_range=(0.1, 30), verbose=0, response_times=(0.1, 5), ccbox=0):
        """
        A time series object

        Parameters
        ----------
        values: array_like
            Acceleration time series - should be in m/s2
        dt: float
            Time step
        label: str (default='m1')
            Used for plotting
        smooth_freq_range: tuple [floats] (default=(0.1, 30))
            lower and upper bound of frequency range to compute the smoothed Fourier amplitude spectrum
        verbose: int (default=0)
            Level of output verbosity
        response_times: tuple of floats (default=(0.1, 5))
        ccbox: int
            colour index for plotting
        """
        super(AccSignal, self).__init__(values, dt, label=label, smooth_freq_range=smooth_freq_range, verbose=verbose, ccbox=ccbox)
        if len(response_times) == 2:
            self.response_times = np.linspace(response_times[0], response_times[1], 100)
        else:
            self.response_times = np.array(response_times)
        self._velocity = np.zeros(self.npts)
        self._displacement = np.zeros(self.npts)
        self._cached_params = {}
        self._cached_response_spectra = False
        self._cached_disp_and_velo = False
        self._cached_xi = 0.05
        self._s_a = None
        self._s_v = None
        self._s_d = None

    def clear_cache(self):
        self._cached_smooth_fa = False
        self._cached_fa = False
        self._cached_response_spectra = False
        self._cached_disp_and_velo = False
        self.reset_all_motion_stats()

    def generate_response_spectrum(self, response_times=None, xi=-1, min_dt_ratio=4):
        """
        Generate the response spectrum for the response_times for a given
        damping (xi). default xi = 0.05
        """
        if self.verbose:
            print('Generating response spectra')
        if response_times is not None:
            self.response_times = response_times
        if self.response_times[0] != 0:
            min_non_zero_period = self.response_times[0]
        else:
            min_non_zero_period = self.response_times[1]
        target_dt = max(min_non_zero_period / 20, self.dt / min_dt_ratio)  # limit to ratio of motion time step
        if target_dt < self.dt:
            values_interp, dt_interp = interp_array_to_approx_dt(self.values, self.dt, target_dt, even=False)
        else:
            values_interp = self.values
            dt_interp = self.dt

        if xi == -1:
            xi = self._cached_xi
        self._s_d, self._s_v, self._s_a = dh.pseudo_response_spectra(values_interp, dt_interp, self.response_times, xi)
        self._cached_response_spectra = True

    @property
    def s_a(self):
        """
        Pseudo spectral response acceleration of linear SDOF
        """
        if not self._cached_response_spectra:
            self.generate_response_spectrum()
        return self._s_a

    @property
    def s_v(self):
        """
        Pseudo spectral response velocity of linear SDOF
        """
        if not self._cached_response_spectra:
            self.generate_response_spectrum()
        return self._s_v

    @property
    def s_d(self):
        """
        Spectral response displacement of linear SDOF
        """
        if not self._cached_response_spectra:
            self.generate_response_spectrum()
        return self._s_d

    def correct_me(self):
        """
        This provides a correction to an acceleration time series
        """

        # self.remove_average()
        # self.butter_pass([0.1, 10])

        disp = ss.detrend(self.displacement)
        vel = np.zeros(self.npts)
        acc = np.zeros(self.npts)
        for i in range(self.npts - 1):  # MEANS THAT ACC has a pulse at the end.
            vel[i + 1] = (disp[i + 1] - disp[i]) / self.dt
            acc[i + 1] = (vel[i + 1] - vel[i]) / self.dt
        self.reset_values(acc)

    def remove_rolling_average(self, mtype="velocity", freq_window=5):
        """
        Removes the rolling average

        Parameters
        ----------
        mtype: str
            motion type to apply method to
        freq_window: int
            window for applying the rolling average

        Returns
        -------

        """
        if mtype == "velocity":
            mot = self.velocity
        else:
            mot = self.values
        width = int(1. / (freq_window * self.dt))
        if width < 1:
            raise ValueError("freq_window to high")
        roll = np.zeros_like(mot)
        for i in range(len(mot)):
            if i < width / 2:
                cc = i + int(width / 2) + 1
                roll[i] = np.mean(mot[:cc])
            elif i > len(mot) - width / 2:
                cc = i - int(width / 2)
                roll[i] = np.mean(mot[cc:])
            else:
                cc1 = i - int(width / 2)
                cc2 = i + int(width / 2) + 1
                roll[i] = np.mean(mot[cc1:cc2])
        if mtype == "velocity":
            velocity = self.velocity - roll
            acc = np.diff(velocity) / self.dt
            acc = np.insert(acc, 0, velocity[0] / self.dt)
            self._values = acc
        else:
            self._values -= roll
        self.clear_cache()

    def rebase_displacement(self):
        """
        Correction to make the displacement zero at the end of the record
        """

        end_disp = self.displacement[-1]

        acceleration_correction = 2 * end_disp / (self.dt * self.npts)
        self._values -= acceleration_correction
        self.clear_cache()

    def generate_displacement_and_velocity_series(self, trap=True):
        """
        Calculates the displacement and velocity time series
        """
        self._velocity, self._displacement = sd.velocity_and_displacement_from_acceleration(self.values,
                                                                                          self.dt, trap=trap)
        self._cached_disp_and_velo = True

    @property
    def velocity(self):
        """Velocity time series"""
        if not self._cached_disp_and_velo:
            self.generate_displacement_and_velocity_series()
        return self._velocity

    @property
    def displacement(self):
        """Displacement time series"""
        if not self._cached_disp_and_velo:
            self.generate_displacement_and_velocity_series()
        return self._displacement

    def generate_peak_values(self):
        """
        Determines the peak signal values
        """
        deprecation("generate_peak_values() is no longer in use, all peak values are lazy loaded.")

    @property
    def pga(self):
        """Absolute peak ground acceleration"""
        if "pga" in self._cached_params:
            return self._cached_params["pga"]
        else:
            pga = sm.calc_peak(self.values)
            self._cached_params["pga"] = pga
            return pga

    @property
    def pgv(self):
        """Absolute peak ground velocity"""
        if "pgv" in self._cached_params:
            return self._cached_params["pgv"]
        else:
            pgv = sm.calc_peak(self.velocity)
            self._cached_params["pgv"] = pgv
            return pgv

    @property
    def pgd(self):
        """Absolute peak ground displacement"""
        if "pgd" in self._cached_params:
            return self._cached_params["pgd"]
        else:
            pgd = sm.calc_peak(self.displacement)
            self._cached_params["pgd"] = pgd
            return pgd

    def generate_duration_stats(self):
        """
        Deprecated: Use eqsig.im.calc_sig_dur or eqsig.im.calc_sig_dur, eqsig.im.calc_brac_dur or esig.im.calc_a_rms
        """
        deprecation("Use eqsig.im.calc_sig_dur or eqsig.im.calc_sig_dur, eqsig.im.calc_brac_dur or esig.im.calc_a_rms")
        abs_motion = abs(self.values)

        time = np.arange(self.npts) * self.dt
        # Bracketed duration
        ind01 = np.where(abs_motion / 9.8 > 0.01)  # 0.01g
        time2 = time[ind01]
        try:
            self.t_b01 = time2[-1] - time2[0]
            # rms acceleration in m/s/s
            self.a_rms01 = np.sqrt(1 / self.t_b01 * np.trapz((self.values[ind01[0][0]:ind01[0][-1]]) ** 2, dx=self.dt))
        except IndexError:
            self.t_b01 = -1.
            self.a_rms01 = -1.

        ind05 = np.where(abs_motion / 9.8 > 0.05)  # 0.05g
        time05 = time[ind05]
        try:
            self.t_b05 = time05[-1] - time05[0]
            self.a_rms05 = np.sqrt(1 / self.t_b05 * np.trapz((self.values[ind05[0][0]:ind05[0][-1]]) ** 2, dx=self.dt))
        except IndexError:
            self.t_b05 = -1.
            self.a_rms05 = -1.
        ind10 = np.where(abs_motion / 9.8 > 0.1)  # 0.10g
        time10 = time[ind10]
        try:
            self.t_b10 = time10[-1] - time10[0]
            self.a_rms10 = np.sqrt(1 / self.t_b10 * np.trapz((self.values[ind10[0][0]:ind10[0][-1]]) ** 2, dx=self.dt))
        except IndexError:
            self.t_b10 = -1.
            self.a_rms10 = -1.

        # Trifunac and Brady
        self.sd_start, self.sd_end = sm.calc_sig_dur_vals(self.values, self.dt, se=True)

        self.t_595 = self.sd_end - self.sd_start

    def generate_cumulative_stats(self):
        """
        Deprecated: Use eqsig.im.calc_arias_intensity, eqsig.im.calc_cav

        """
        deprecation("Use eqsig.im.calc_arias_intensity, eqsig.im.calc_cav")
        # Arias intensity in m/s
        self.arias_intensity_series = im.calc_arias_intensity(self)
        self.arias_intensity = self.arias_intensity_series[-1]
        self.cav_series = im.calc_cav(self)
        self.cav = self.cav_series[-1]

    def generate_all_motion_stats(self):
        """
        Deprecated: Use eqsig.im functions to calculate stats
        """
        deprecation("Use eqsig.im functions to calculate stats")

        self.generate_duration_stats()
        self.generate_cumulative_stats()

    def reset_all_motion_stats(self):
        self.t_b01 = 0.0
        self.t_b05 = 0.0
        self.t_b10 = 0.0
        self.a_rms01 = 0.0
        self.a_rms05 = 0.0
        self.a_rms10 = 0.0
        self.t_595 = 0.0
        self.sd_start = 0.0
        self.sd_end = 0.0
        self.arias_intensity = 0.0
        self._cached_params = {}

    def response_series(self, response_times=None, xi=-1):
        """
        Generate the response time series for a set of elastic SDOFs for a given
        damping (xi).
        """
        if self.verbose:
            print('Generating response series')
        if response_times is not None:
            self.response_times = response_times
        if xi == -1:
            xi = self._cached_xi
        resp_u, resp_v, resp_a = dh.response_series(self.values, self.dt, self.response_times, xi)
        return resp_u, resp_v, resp_a



import numpy as np
import scipy.integrate


def velocity_and_displacement_from_acceleration(acceleration, dt, trap=True):
    """
    Computes the velocity and acceleration of an acceleration time series, using numerical integration.

    Parameters
    ----------
    acceleration: array_like
        acceleration time series
    dt: float
        time step
    trap: bool (default=True)
        if True then uses trapezium integration

    Returns
    -------
    velocity: array_like (len=len(acceleration))
        velocity time series
    displacement: array_like (len=len(acceleration))
        displacement time series
    """

    if trap is False:
        vel_inc = acceleration * dt
        velocity = np.cumsum(vel_inc)
        np.insert(velocity, 0, 0)
        velocity = velocity[:-1]
        disp_inc = velocity * dt
        displacement = np.cumsum(disp_inc)
        np.insert(displacement, 0, 0)
        displacement = displacement[:-1]
    else:

        velocity = scipy.integrate.cumtrapz(acceleration, dx=dt, initial=0)
        displacement = scipy.integrate.cumtrapz(velocity, dx=dt, initial=0)

    return velocity, displacement



from eqsig.single import Signal, AccSignal
from eqsig import design_spectra
from eqsig.functions import *
from eqsig.multiple import Cluster, combine_at_angle, compute_rotated
from eqsig.im import calc_significant_duration, calculate_peak  # deprecated load
from eqsig import im as measures
from eqsig import im
from eqsig import stockwell
from eqsig.loader import save_signal, load_signal, save_values_and_dt, load_values_and_dt, load_asig, load_sig
from eqsig import loader
from eqsig import __about__
from eqsig import sdof

__project__ = __about__.__project__
__author__ = __about__.__author__
__version__ = __about__.__version__
__license__ = __about__.__license__



import numpy as np
import matplotlib.pyplot as plt

from eqsig.single import AccSignal
from tests.conftest import TEST_DATA_DIR


def show_test_motion():
    record_path = TEST_DATA_DIR
    record_filename = 'test_motion_dt0p01.txt'
    motion_step = 0.01
    rec = np.loadtxt(record_path + record_filename)
    acc_signal = AccSignal(rec, motion_step)
    acc_signal.generate_displacement_and_velocity_series()
    bf, sp = plt.subplots(3)
    sp[0].plot(acc_signal.time, acc_signal.values)
    sp[1].plot(acc_signal.time, acc_signal.velocity)
    sp[2].plot(acc_signal.time, acc_signal.displacement)
    plt.show()


def show_response_spectra_at_high_frequencies():
    record_path = TEST_DATA_DIR
    test_filename = 'test_motion_true_spectra_acc.csv'
    data = np.loadtxt(record_path + test_filename, skiprows=1, delimiter=",")
    times = data[:40, 0]
    ss_s_a = data[:40, 1]

    record_filename = 'test_motion_dt0p01.txt'
    motion_step = 0.01
    rec = np.loadtxt(record_path + record_filename)
    # acc_signal = AccSignal(rec, motion_step, response_times=times)
    # s_a = acc_signal.s_a
    #
    # a_times = acc_signal.response_times
    # s_d, s_v, s_a = dh.pseudo_response_spectra(rec, motion_step, times, xi=0.05)
    # s_d, s_v, s_a = dh.true_response_spectra(rec, motion_step, times, xi=0.05)
    acc_signal = AccSignal(rec, motion_step, response_times=times)
    s_a = acc_signal.s_a

    s_a_in_g = s_a / 9.81

    # srss1 = sum(abs(s_a_in_g - ss_s_a))
    plt.plot(times, s_a_in_g, label="eqsig")
    plt.plot(times, ss_s_a, label="true-ish")
    plt.legend()
    plt.show()


def show_fourier_spectra_stable_against_aliasing():
    record_path = TEST_DATA_DIR

    record_filename = 'test_motion_dt0p01.txt'
    motion_step = 0.01
    rec = np.loadtxt(record_path + record_filename)
    rec2 = np.zeros(2 ** 13)
    rec2[:len(rec)] = rec
    org_signal = AccSignal(rec, motion_step)
    extended_signal = AccSignal(rec2, motion_step)

    rec_split = []
    for i in range(int(len(rec2) / 2)):
        rec_split.append(rec2[i * 2])

    acc_split = AccSignal(rec_split, motion_step * 2)

    bf, sp = plt.subplots(2)
    sp[0].plot(org_signal.time, org_signal.values)
    sp[0].plot(extended_signal.time, extended_signal.values)
    sp[0].plot(acc_split.time, acc_split.values)

    sp[1].plot(org_signal.fa_frequencies, abs(org_signal.fa_spectrum), lw=0.7, label="original")
    sp[1].plot(acc_split.fa_frequencies, abs(acc_split.fa_spectrum), lw=0.7, label="split")
    sp[1].plot(extended_signal.fa_frequencies, abs(extended_signal.fa_spectrum), lw=0.7, label="full")

    plt.legend()
    plt.show()


import numpy as np
from eqsig.single import Signal, AccSignal


def load_values_and_dt(ffp):
    """
    Loads values and time step that were saved in eqsig input format.

    Parameters
    ----------
    ffp: str
        Full file path to output file

    Returns
    -------
    values: array_like
        An array of values
    dt: float
        Time step

        """
    data = np.genfromtxt(ffp, skip_header=1, delimiter=",", names=True, usecols=0)
    dt = data.dtype.names[0].split("_")[-1]
    dt = "." + dt[1:]
    dt = float(dt)
    values = data.astype(np.float)
    return values, dt


# def fast_nga_loader(ffp):
#     data = np.genfromtxt(ffp, skip_header=4, names=True)
#     data.flatten()
#     dt = data.dtype.names[0].split("DT=")[-1]
#     dt = "." + dt[1:]
#     print(dt)
#     dt = float(dt)
#     values = data.astype(np.float)
#     return values, dt


def save_values_and_dt(ffp, values, dt, label):
    """
    Exports acceleration values to the eqsig format.

    Parameters
    ----------
    ffp: str
        Full file path to output file
    values: array_like
        An array of values
    dt: float
        Time step
    label: str
        A label of the data

    Returns
    -------

    """
    para = [label, "%i %.4f" % (len(values), dt)]
    for i in range(len(values)):
        para.append("%.6f" % values[i])
    ofile = open(ffp, "w")
    ofile.write("\n".join(para))
    ofile.close()


def load_signal(ffp, astype='sig'):
    vals, dt = load_values_and_dt(ffp)
    if astype == "signal":
        return Signal(vals, dt)
    elif astype == "acc_sig":
        return AccSignal(vals, dt)


def load_sig(ffp):
    vals, dt = load_values_and_dt(ffp)
    return Signal(vals, dt)


def load_asig(ffp):
    vals, dt = load_values_and_dt(ffp)
    return AccSignal(vals, dt)


def save_signal(ffp, signal):
    save_values_and_dt(ffp, signal.values, signal.dt, signal.label)


def load_3_comp_values_and_dt_from_v2a(ffp):
    """
    Loads a ground motion file stored in the V2A format

    Parameters
    ----------
    ffp

    Returns
    -------

    """
    a = open(ffp)
    b = a.readlines()
    a.close()
    npts = None
    dt = None
    for line in b:
        if 'Number of points' in line:
            parts = line.split('points')[1]
            n_str = parts.split('Duration')[0]
            npts = int(n_str)
        if 'corrected data at' in line:
            parts = line.split('data at')[1]
            dt_str = parts.split('sec')[0]
            dt = float(dt_str)
        if npts is not None and dt is not None:
            break
    b = b[26:]
    elines = int(np.ceil(float(npts) / 10))
    accs = []
    for i in range(3):
        accs.append([])
        print(3 * i * elines, (3 * i + 1) * elines)
        for j in range(3 * i * elines + i * 26, (3 * i + 1) * elines + i * 26):
            accs[i] += b[j].split()
    accs = np.array(accs).astype(float) / 1e3

    return accs[0], accs[1], accs[2], dt

